from argparse import ArgumentParser
from collections import Counter
import pandas as pd
import numpy as np
import torch
import torch.nn as nn
import torch.nn.functional as F
from torch.utils.data import TensorDataset, DataLoader
import pytorch_lightning as pl
from transformers import AutoTokenizer, AutoModel
from transformers.optimization import AdamW


def mk_tensors(txt, tokenizer, max_seq_length):
    tok_res = tokenizer(
        txt, truncation=True, padding="max_length", max_length=max_seq_length
    )
    input_ids = tok_res["input_ids"]
    attention_mask = tok_res["attention_mask"]
    input_ids = torch.tensor(input_ids)
    attention_mask = torch.tensor(attention_mask)
    return input_ids, attention_mask


def mk_ds(txt, tokenizer, max_seq_length, ys):
    input_ids, attention_mask = mk_tensors(txt, tokenizer, max_seq_length)
    return TensorDataset(input_ids, attention_mask, torch.tensor(ys))


class PCDataModule(pl.LightningDataModule):
    def __init__(
        self,
        model_name_or_path,
        max_seq_length,
        min_products_for_category,
        train_batch_size,
        val_batch_size,
        dataloader_num_workers,
        pin_memory,
        data_file_path=None,
        dataframe=None,
    ):
        super().__init__()
        self.data_file_path = data_file_path
        self.dataframe = dataframe
        self.min_products_for_category = min_products_for_category
        self.model_name_or_path = model_name_or_path
        self.tokenizer = AutoTokenizer.from_pretrained(self.model_name_or_path)
        self.max_seq_length = max_seq_length
        self.train_batch_size = train_batch_size
        self.val_batch_size = val_batch_size
        self.dataloader_num_workers = dataloader_num_workers
        self.pin_memory = pin_memory
        self.num_classes = None

    def setup(self, stage=None):
        if self.dataframe is None:
            self.dataframe = pd.read_csv(self.data_file_path)
            self.dataframe.dropna(inplace=True)

        cats = self.dataframe.category.apply(lambda x: x.split("|"))
        cat2cnt = Counter((j for i in cats for j in i))
        i2cat = sorted(
            k for k, v in cat2cnt.items() if v > self.min_products_for_category
        )
        cat2i = {v: k for k, v in enumerate(i2cat)}
        self.num_classes = len(i2cat)
        self.i2cat, self.cat2i = i2cat, cat2i

        ys = np.zeros((len(self.dataframe), len(i2cat)))
        for i, cats in enumerate(self.dataframe.category):
            idx_pos = [cat2i[cat] for cat in cats.split("|") if cat in cat2i]
            ys[i, idx_pos] = 1

        msk_val = self.dataframe.is_validation == 1
        self.df_trn = self.dataframe[~msk_val]
        self.df_val = self.dataframe[msk_val]
        idx_trn = np.where(~msk_val)[0]
        idx_val = np.where(msk_val)[0]
        self.ys_trn, self.ys_val = ys[idx_trn], ys[idx_val]

        self.train_dataset = mk_ds(
            list(self.df_trn.title), self.tokenizer, self.max_seq_length, self.ys_trn
        )
        self.eval_dataset = mk_ds(
            list(self.df_val.title), self.tokenizer, self.max_seq_length, self.ys_val
        )

    def train_dataloader(self):
        return DataLoader(
            self.train_dataset,
            batch_size=self.train_batch_size,
            num_workers=self.dataloader_num_workers,
            pin_memory=self.pin_memory,
        )

    def val_dataloader(self):
        return DataLoader(
            self.eval_dataset,
            batch_size=self.val_batch_size,
            num_workers=self.dataloader_num_workers,
            pin_memory=self.pin_memory,
        )


def getaccu(logits, ys):
    return ((logits > 0.0).int() == ys).float().mean()


class PCModel(pl.LightningModule):
    def __init__(
        self,
        model_name_or_path,
        freeze_bert,
        num_classes,
        learning_rate,
        adam_beta1,
        adam_beta2,
        adam_epsilon,
    ):
        super().__init__()
        self.save_hyperparameters()
        self.model_name_or_path = model_name_or_path
        self.bert = AutoModel.from_pretrained(self.model_name_or_path)
        self.freeze_bert = freeze_bert
        if self.freeze_bert == True:
            for param in self.bert.parameters():
                param.requires_grad = False
        self.num_classes = num_classes
        self.W = nn.Linear(self.bert.config.hidden_size, self.num_classes)

    def forward(self, input_ids, attention_mask):
        h = self.bert(input_ids, attention_mask)["last_hidden_state"]
        h_cls = h[:, 0]
        return self.W(h_cls)

    def training_step(self, batch, batch_idx):
        input_ids, attention_mask, ys = batch
        logits = self(input_ids, attention_mask)
        loss = F.binary_cross_entropy_with_logits(logits, ys)
        accu = getaccu(logits, ys)
        self.log("train_loss", loss, on_epoch=True)
        self.log("train_accu", accu, on_epoch=True)
        return loss

    def validation_step(self, batch, batch_idx):
        input_ids, attention_mask, ys = batch
        logits = self(input_ids, attention_mask)
        loss = F.binary_cross_entropy_with_logits(logits, ys)
        accu = getaccu(logits, ys)
        self.log("valid_loss", loss, on_step=False, sync_dist=True)
        self.log("valid_accu", accu, on_step=False, sync_dist=True)

    def configure_optimizers(self):
        optimizer = AdamW(
            self.parameters(),
            self.hparams.learning_rate,
            betas=(self.hparams.adam_beta1, self.hparams.adam_beta2),
            eps=self.hparams.adam_epsilon,
        )
        return optimizer

    @staticmethod
    def add_model_specific_args(parent_parser):
        parser = ArgumentParser(parents=[parent_parser], add_help=False)
        parser.add_argument(
            "--learning_rate", type=float, help="Learning Rate", default=5e-5
        )
        parser.add_argument(
            "--adam_beta1",
            type=float,
            help="The beta1 parameter in Adam, which is the exponential decay rate for the 1st momentum estimates.",
            default=0.9,
        )
        parser.add_argument(
            "--adam_beta2",
            type=float,
            help="The beta2 parameter in Adam, which is the exponential decay rate for the 2nd momentum estimates.",
            default=0.999,
        )
        parser.add_argument(
            "--adam_epsilon",
            type=float,
            help="The epsilon parameter in Adam, which is a small constant for numerical stability.",
            default=1e-8,
        )
        return parser



from argparse import ArgumentParser
import pytorch_lightning as pl
from pytorch_lightning.callbacks.early_stopping import EarlyStopping
from product_category.pc_model import PCDataModule, PCModel
import os

os.environ["TOKENIZERS_PARALLELISM"] = "false"


def cli_main():

    # ------------
    # args
    # ------------
    parser = ArgumentParser()
    parser.add_argument(
        "--model_name_or_path",
        type=str,
        help="Path to pretrained model or model identifier from huggingface.co/models.",
        default="distilbert-base-cased",
    )
    parser.add_argument(
        "--transfer_learn",
        help="Wether to use transfer learning based on a pretrained model.",
        action="store_true",
    )
    parser.add_argument(
        "--trained_model_path",
        type=str,
        help=("Model used to predict."),
        default="../data/transformer_20210307D3.ckpt",
    )

    parser.add_argument(
        "--data_file_path",
        help="Path to training data file. "
        "Data file should be csv with 3 columns: category (categories separated by '|'),title (str),is_validation (0 or 1). "
        "e.g.: Sports & Outdoors|Outdoor Recreation|Cycling|Clothing|Men|Shorts,Louis Garneau Men's Neo Power Motion Bike Shorts,1",
        type=str,
    )
    parser.add_argument(
        "--freeze_bert",
        help="Whether to freeze the pretrained model.",
        action="store_true",
    )
    parser.add_argument(
        "--max_seq_length",
        help="The maximum total input sequence length after tokenization. Sequences longer "
        "than this will be truncated, sequences shorter will be padded.",
        type=int,
        default=128,
    )
    parser.add_argument(
        "--min_products_for_category",
        help="Minimum number of products for a category to be considered in the model.",
        type=int,
        default=100,
    )
    parser.add_argument(
        "--train_batch_size",
        help="How many samples per batch to load for train dataloader.",
        type=int,
        default=128,
    )
    parser.add_argument(
        "--val_batch_size",
        help="How many samples per batch to load for validation dataloader.",
        type=int,
        default=256,
    )
    parser.add_argument(
        "--dataloader_num_workers",
        help="How many subprocesses to use for data loading. 0 means that the data will be loaded in the main process.",
        type=int,
        default=8,
    )
    parser.add_argument(
        "--pin_memory",
        help="Wether to use pin_memory in pytorch dataloader. If True, the data loader will copy Tensors into CUDA pinned memory before returning them.",
        action="store_true",
    )

    parser = pl.Trainer.add_argparse_args(parser)
    parser = PCModel.add_model_specific_args(parser)
    args = parser.parse_args()

    # ------------
    # data
    # ------------
    data_module = PCDataModule(
        model_name_or_path=args.model_name_or_path,
        data_file_path=args.data_file_path,
        min_products_for_category=args.min_products_for_category,
        max_seq_length=args.max_seq_length,
        train_batch_size=args.train_batch_size,
        val_batch_size=args.val_batch_size,
        dataloader_num_workers=args.dataloader_num_workers,
        pin_memory=args.pin_memory,
    )
    data_module.prepare_data()
    data_module.setup()

    # ------------
    # model
    # ------------
    pcmodel = PCModel(
        model_name_or_path=args.model_name_or_path,
        freeze_bert=args.freeze_bert,
        num_classes=data_module.num_classes,
        learning_rate=args.learning_rate,
        adam_beta1=args.adam_beta1,
        adam_beta2=args.adam_beta2,
        adam_epsilon=args.adam_epsilon,
    )
    if args.transfer_learn:
        pretrained_model = PCModel.load_from_checkpoint(args.trained_model_path)
        pcmodel.bert = pretrained_model.bert
    pcmodel.prepare_data()
    # ------------
    # training
    # ------------
    trainer = pl.Trainer.from_argparse_args(
        args,
        callbacks=[EarlyStopping(monitor="valid_loss")],
        stochastic_weight_avg=True,
    )
    trainer.fit(pcmodel, data_module)


if __name__ == "__main__":
    cli_main()



from argparse import ArgumentParser
from transformers import AutoTokenizer
from product_category.pc_model import PCModel, mk_tensors
import torch
import numpy as np


def predict(txt, pcmodel, tokenizer):
    input_ids, attention_mask = mk_tensors([txt], tokenizer, 128)
    logits = pcmodel(input_ids, attention_mask)[0]
    scores = torch.sigmoid(logits)
    return scores.detach().numpy()


def cli_main():
    parser = ArgumentParser()
    parser.add_argument(
        "-t", "--text", type=str, help="Product info text to predict.", required=True
    )
    parser.add_argument(
        "--trained_model_path",
        type=str,
        help="Model used to predict.",
        default="../data/transformer_20210307D3.ckpt",
    )
    parser.add_argument(
        "--i2cat_path",
        type=str,
        help="File name for the ordered list of categories. Each line for a category.",
        default="../data/i2cat_transformer_20210307D3.txt",
    )
    parser.add_argument(
        "--tokenizer_name",
        type=str,
        help="Tokenizer name.",
        default="distilbert-base-cased",
    )

    parser.add_argument(
        "--topn",
        type=int,
        help="Number of top predicted categories to display.",
        default=5,
    )

    args = parser.parse_args()
    with open(args.i2cat_path) as f:
        i2cat = f.read().splitlines()

    pcmodel = PCModel.load_from_checkpoint(args.trained_model_path)
    pcmodel.eval()
    tokenizer = AutoTokenizer.from_pretrained(args.tokenizer_name)
    scores = predict(args.text, pcmodel, tokenizer)
    top_scores = sorted(scores, reverse=True)[: args.topn]
    top_icats = np.argsort(-scores)[: args.topn]
    for i, scr in zip(top_icats, top_scores):
        print(f"{i2cat[i]}: {scr:.3f}")


if __name__ == "__main__":
    cli_main()
